"""
Training interfaces are stored here
"""
import functools
from dataclasses import dataclass
from typing import Optional, Callable, Tuple, List, Dict, Any, Union
from abc import ABC, abstractmethod

import torch
import csv
import os
import pandas as pd
import threading
import concurrent.futures as futures

from tinycss2 import tokenizer
from torch import nn
from torch.utils import data
from concurrent.futures import ThreadPoolExecutor, Future

from datasets import DatasetDict, Dataset
from transformers import PreTrainedTokenizer

from .base import TensorTree, parallel_pytree_map
from .grad_utils import AbstractGradientControl, AutorescaleGradientControl
from .losses import MainLossInterface, MemAccessLossInterface
from .model import CausalLMTrainer, Logger, CausalLMCore
from .data.pretraining import PretrainingLoaderConfig, create_dataloader_factory

try:
    from IPython.display import clear_output

    CLEAR_OUTPUT_AVAILABLE = True
except ImportError:
    CLEAR_OUTPUT_AVAILABLE = False


# Define more concrete implementatons.

@dataclass
class TrainingConfig:
    """
    A central dataclass that holds all the dynamic features needed to
    properly set up and run a training instance.

    Fields:
    --------
    loader_factories: Callable[[int], Dict[str, data.DataLoader]]
        A function that produces data loaders for training and validation,
        keyed by dataset names. Takes the worker num as input
    training_run_prefix: str
        Prefix for naming or identifying this training run.
    metrics_logging_directory: str
        Directory where metrics logs will be saved.
    checkpoint_save_directory: str
        Directory where model checkpoints will be saved.
    checkpoint_batch_frequency: int
        Frequency (in batches) at which checkpoints are saved.
    num_workers: int
        Number of workers used for data loading.
    num_epochs: int
        Total number of epochs for the training run.
    epoch_position: Optional[int]
        Current epoch position in the training process. Useful for resuming
        from checkpoints. Defaults to None if starting from the beginning.
    """
    @property
    def num_workers(self)->int:
        return len(self.devices)

    # devices
    devices: List[torch.device]

    # Loader config
    loader_config: PretrainingLoaderConfig

    # Training and logging niceties
    training_run_prefix: str
    metrics_logging_directory: str
    checkpoint_save_directory: str
    checkpoint_batch_frequency: int

    # Important features
    num_workers: int
    num_epochs: int
    epoch_position: Optional[int] = None


class LoggingContext:
    """
    Context manager to provide a thread pool executor for logging or other
    asynchronous tasks.
    """

    def __init__(self, max_workers: int = 2):
        self.max_workers = max_workers
        self.executor = None

    def __enter__(self):
        self.executer = ThreadPoolExecutor(max_workers=self.max_workers)
        return self.executor

    def __exit__(self, exc_type, exc_value, traceback):
        self.executor.shutdown(wait=True)


class LogMetrics:
    def __init__(self, training_config: TrainingConfig):
        """
        A logging interface for saving metrics to an in-memory DataFrame and
        periodically writing epochs to a CSV file.
        """
        self.file = training_config.metrics_logging_directory
        self.current_epoch = training_config.epoch_position
        self.data = None  # DataFrame initialized on first log call if not resuming
        self._initialized = training_config.epoch_position > 0

    def make_logging_callback(self,
                              worker: int,
                              epoch: int,
                              batch: int,
                              pass_type: str) -> Callable[[Dict[str, Any]], None]:
        """
        Creates a metric callback that is bound to log information under the
        given worker, epoch, and batch.

        :param worker: The worker id.
        :param epoch: The epoch.
        :param batch: The batch in the epoch.
        :param pass_type: Training, validation, or test.
        :return: The functional callback.
        """

        def logging_callback(metrics: Dict[str, Any]) -> None:
            self.log(epoch, batch, worker, metrics, pass_type)

        return logging_callback

    def log(self,
            epoch: int,
            batch: int,
            worker: int,
            pass_type: str,
            metrics: Dict[str, Any]):
        """
        Logs metrics for a given epoch, batch, and worker to an in-memory
        DataFrame.

        :param epoch: Current epoch number.
        :param batch: Current batch number.
        :param worker: Worker ID or number.
        :param pass_type: Training, validation, or test.
        :param metrics: Dictionary of metric names and their values.
        """
        # Initialize the DataFrame if it hasn't been set up yet
        if self.data is None:
            columns = ["epoch", "batch", "worker", "pass_type"] + list(metrics.keys())
            self.data = pd.DataFrame(columns=columns)
        metrics = {name: float(metric) if isinstance(metric, torch.Tensor) else metric
                   for name, metric in metrics.items()}
        # Append the new row of metrics
        row = {"epoch": epoch, "batch": batch, "worker": worker, "pass_type": pass_type, **metrics}
        self.data = pd.concat([self.data, pd.DataFrame([row])], ignore_index=True)

    def write_epoch(self, ) -> Dict[str, str]:
        """
        Writes the accumulated data for the current epoch to the CSV file and
        returns a dictionary with formatted strings of average metrics, organized
        by pass type (e.g., training, validation).

        :return: A dictionary with pass type keys and formatted average summaries.
        """
        if self.data is None or self.data.empty:
            return {f"No data available for epoch {self.current_epoch}."}

        # Group data by pass_type and calculate averages for each group
        summaries = {}
        for pass_type, group in self.data.groupby("pass_type"):
            # Select and average only the specified summary columns
            summary_data = group.select_dtypes(include=['number'])
            averages = summary_data.mean().to_dict()

            # Format the averages into a readable string for this pass_type
            summary_str = f"Epoch {self.current_epoch}, {pass_type} averages: " + ", ".join(
                f"average {metric}: {value:.4f}" for metric, value in averages.items()
            )
            summaries[pass_type] = summary_str

        # Write the epoch data to the CSV file using pandas
        self.data.to_csv(self.file, mode='a' if self._initialized else 'w',
                         index=False, header=not self._initialized)
        self._initialized = True  # File is now initialized

        # Clear data for the current epoch and reinitialize the DataFrame
        self.data = pd.DataFrame(columns=self.data.columns)
        self.current_epoch += 1  # Increment to the next epoch

        return summaries


class TerminalDisplay:
    """
    A helper class that initializes and keeps
    track of various worker instances and whatever
    feedback they are providing.

    """

    def __init__(self, training_config: TrainingConfig):
        """
        Initializes a terminal display for a certain number
        of workers. Each worker gets a line, and we use
        carriage returns to overwrite when seeing updates
        :param num_workers: The number of workers to reserver
        space for.
        """
        self.num_workers = training_config.num_workers
        self.metrics_status = []
        self.worker_display_status = ["No status yet"] * training_config.num_workers
        self.render()

    def make_terminal_callback(self,
                               worker: int,
                               epoch: int,
                               batch: int,
                               ) -> Callable[[str], Any]:
        """
        Creates a terminal callback bound to a particular worker, epoch, and batch
        :param worker: The worker to bind to
        :param epoch: The epoch to display as
        :param batch: The batch to display as
        :return: A callback that will invoke a terminal display
        """

        def terminal_callback(message: str):
            msg = f"W: {worker} | E: {epoch} | B: {batch} | Feedback" + message
            self.worker_display_status[worker] = msg
            self.render()

        return terminal_callback

    def store_epoch_message(self, message: str):
        """
        Stores metrics related to an entire epoch
        """
        self.metrics_status.append(message)
        self.render()

    def render(self):
        """
        Renders the worker status to the terminal
        """
        if CLEAR_OUTPUT_AVAILABLE:
            clear_output()

        output = "----Epoch metrics -----\n"
        output += "\n".join(self.metrics_status)
        output += "----worker feedback----\n"
        output += "\n".join(self.worker_display_status)
        print(output)


class CheckpointProcess:
    """
    A model checkpoint process, that
    can be utilized for training.

    It is capable of saving the model
    to a given location every so often.
    """

    def __init__(self, model: CausalLMCore, training_config: TrainingConfig):
        self.model = model

        self.epoch = training_config.epoch_position
        self.folder_path = training_config.checkpoint_save_directory
        self.checkpoint_every_n_batches = training_config.checkpoint_batch_frequency
        self.batch = 0
        self.prefix = training_config.training_run_prefix

        os.makedirs(self.folder_path, exist_ok=True)

    def save_checkpoint(self):
        name = self.prefix + "_" + f"epoch_{self.epoch}_batch_{self.batch}"
        path = os.path.join(self.folder_path, name)
        self.model.save_to_folder(path)

    def step_batch(self):
        if self.batch % self.checkpoint_every_n_batches == 0:
            self.save_checkpoint()
        self.batch += 1

    def step_epoch(self):
        self.save_checkpoint()
        self.batch = 0
        self.epoch += 1


@dataclass
class TrainingResources:
    """
    Contains the resources used
    to log and display feedback
    while training the model
    """
    # Device
    device: torch.device
    num_workers: int

    # Logging pieces
    metrics_logger: LogMetrics
    terminal_display: TerminalDisplay
    logging_thread: ThreadPoolExecutor

    # Checkpointing and Optim
    core: CausalLMCore
    trainer: CausalLMTrainer
    checkpointing: CheckpointProcess
    optim: torch.optim.Optimizer


def run_training_epoch(
        worker_num: int,
        epoch_num: int,
        train_loader: data.DataLoader,
        training_resources: TrainingResources,
):
    """
    Runs a singlular training epoch. This includes taking the
    loaders, transferring the relevant bits, advancing optim,
    etc.
    :param worker_num: The worker num associated with this. Used for logging
    :param epoch_num: The epoch num associated with this. Used for logging
    :param train_loader: The training dataloader
    :param training_resources: Various training resources. See the class
    """

    # Run training pass
    for i, (tokens, targets, nonpadding_mask) in train_loader:

        # If main process, advance checkpointing.
        if worker_num == 0:
            # The first worker is in charge of checkpointing
            training_resources.checkpointing.step_batch()

        # Move all to the right device
        tokens = tokens.to(device=training_resources.device)
        targets = targets.to(device=training_resources.device)
        nonpadding_mask = nonpadding_mask.to(device=training_resources.device)

        # Compute the scaling factor. This is used to perform an average over all the active tokens
        scaling_factor = (nonpadding_mask).sum().to(training_resources.core.dtype)
        scaling_factor = 1 / scaling_factor
        scaling_factor = float(scaling_factor)

        # Setup the logging and feedback
        terminal_callback = training_resources.terminal_display.make_terminal_callback(worker_num,
                                                                                       epoch_num,
                                                                                       i)
        metrics_callback = training_resources.metrics_logger.make_logging_callback(worker_num,
                                                                                   epoch_num,
                                                                                   i,
                                                                                   "training_pass")
        logging_case = Logger(training_resources.logging_thread, terminal_callback, metrics_callback)

        # Perform the training step. Most of your time is spent here.
        training_resources.trainer.step(tokens, targets, ~nonpadding_mask, logging_case,
                                        scheduling_rates=(scaling_factor, scaling_factor))

        # Fetch the gradients off the model. Perform our all reduce. Then integrate
        # the results back into the model.

        def reduce_gradients(parameter: torch.Tensor):
            if parameter.grad is not None:
                grad = torch.distributed.all_reduce(parameter.grad, op=torch.distributed.ReduceOp.SUM)
                grad = grad / training_resources.num_workers
                parameter.grad = grad

        parallel_pytree_map(reduce_gradients, training_resources.core.parameters())

        # Perform the optim step
        training_resources.optim.step()
        training_resources.optim.zero_grad()


def run_validation_epoch(
        worker_num: int,
        epoch_num: int,
        validation_loader: data.DataLoader,
        training_resources: TrainingResources,
):
    """
    Runs a validation epoch process
    :param worker_num: The worker num associated with this. Used for logging
    :param epoch_num: The epoch num associated with this. Used for logging
    :param validation_loader: The validation dataloader
    :param training_resources: Various training resources. See the class
    """
    # Run training pass
    for i, (tokens, targets, nonpadding_mask) in validation_loader:
        # Move all to the right device
        tokens = tokens.to(device=training_resources.device)
        targets = targets.to(device=training_resources.device)
        nonpadding_mask = nonpadding_mask.to(device=training_resources.device)

        # Compute the scaling factor. This is used to perform an average over all the active tokens
        scaling_factor = (nonpadding_mask).sum().to(training_resources.core.dtype)
        scaling_factor = 1 / scaling_factor
        scaling_factor = float(scaling_factor)

        # Setup the logging and feedback
        terminal_callback = training_resources.terminal_display.make_terminal_callback(worker_num,
                                                                                       epoch_num,
                                                                                       i)
        metrics_callback = training_resources.metrics_logger.make_logging_callback(worker_num,
                                                                                   epoch_num,
                                                                                   i,
                                                                                   "validation_pass")
        logging_case = Logger(training_resources.logging_thread, terminal_callback, metrics_callback)

        # Perform the training step. Most of your time is spent here.
        training_resources.trainer.step(tokens, targets, ~nonpadding_mask, logging_case,
                                        scheduling_rates=(scaling_factor, scaling_factor))

        # Zero the grads. We do not need them. But we cannot let them get full and give us NAN's
        training_resources.optim.zero_grad()


def run_test_epoch(
        worker_num: int,
        epoch_num: int,
        test_loader: data.DataLoader,
        training_resources: TrainingResources,
):
    """
    Runs a validation epoch process
    :param worker_num: The worker num associated with this. Used for logging
    :param epoch_num: The epoch num associated with this. Used for logging
    :param test_loader: The validation dataloader
    :param training_resources: Various training resources. See the class
    """
    # Run training pass
    for i, (tokens, targets, nonpadding_mask) in test_loader:
        # Move all to the right device
        tokens = tokens.to(device=training_resources.device)
        targets = targets.to(device=training_resources.device)
        nonpadding_mask = nonpadding_mask.to(device=training_resources.device)

        # Compute the scaling factor. This is used to perform an average over all the active tokens
        scaling_factor = (nonpadding_mask).sum().to(training_resources.core.dtype)
        scaling_factor = 1 / scaling_factor
        scaling_factor = float(scaling_factor)

        # Setup the logging and feedback
        terminal_callback = training_resources.terminal_display.make_terminal_callback(worker_num,
                                                                                       epoch_num,
                                                                                       i)
        metrics_callback = training_resources.metrics_logger.make_logging_callback(worker_num,
                                                                                   epoch_num,
                                                                                   i,
                                                                                   "test_pass")
        logging_case = Logger(training_resources.logging_thread, terminal_callback, metrics_callback)

        # Perform the training step. Most of your time is spent here.
        training_resources.trainer.step(tokens, targets, ~nonpadding_mask, logging_case,
                                        scheduling_rates=(scaling_factor, scaling_factor))

        # Zero the grads. We do not need them. But we cannot let them get full and give us NAN's
        training_resources.optim.zero_grad()


def run_training_process(worker_num: int,
                         training_configs: List[TrainingConfig],
                         loader_factory: Callable[[int], data.DataLoader],
                         model_factory: Callable[[torch.device], Tuple[CausalLMCore, CausalLMTrainer]],
                         optim_factory: Callable[[nn.Module], torch.optim.Optimizer],
                         schedule_factory: Callable[
                             [torch.optim.Optimizer], Optional[torch.optim.lr_scheduler.LRScheduler]],
                         ):
    """
    Runs a training process, in parallel, and presumably
    on the GPU.

    :param worker_num: The worker number assigned. Given out by spawn
    :param training_configs: The training configs to use. Each will be used in sequence, advancing from the first
                             batch length.
    :param loader_factory: A factory that when given a rank returns a working loader set
    :param model_factory: A factory method capable of making a model given a device
    :param optim_factory: A factory method capable of providing an optim when given a model.
    :param schedule_factory: A factory method capable of making a schedule given an optimizer.
    """
    # Setup the models and training mechanisms
    device = training_configs[0].devices[worker_num]
    core, trainer = model_factory(device)
    optim = optim_factory(trainer)
    checkpointing = CheckpointProcess(core, training_configs)
    schedule = schedule_factory(optim)
    epoch_num = training_configs[0].epoch_position
    num_workers = training_configs[0].num_workers

    with ThreadPoolExecutor(max_workers=2) as logging_threads:
        # Run training under configurations
        for training_config in training_configs:
            training_config.epoch_position = epoch_num

            terminal_display = TerminalDisplay(training_config)
            metrics_logger = LogMetrics(training_config)

            # Create the training resources
            resources = TrainingResources(
                device=device,
                num_workers=num_workers,
                logging_thread=logging_threads,
                terminal_display=terminal_display,
                metrics_logger=metrics_logger,
                core=core,
                trainer=trainer,
                checkpointing=checkpointing,
                optim=optim,
            )

            # Create the dataloaders
            loaders = loader_factory(worker_num)

            for _ in range(training_config.num_epochs):
                # Run primary epochs
                run_training_epoch(worker_num, epoch_num, loaders["train_loader"], resources)
                run_validation_epoch(worker_num, epoch_num, loaders["validation_loader"], resources)

                # Perform end of epoch processes, such as checkpointing and advancement
                resources.checkpointing.step_epoch()
                summaries = resources.metrics_logger.write_epoch()
                resources.terminal_display.store_epoch_message("during training: " + summaries["training_pass"])
                resources.terminal_display.store_epoch_message("during validation: " + summaries["validation_pass"])
                if schedule is not None:
                    schedule.step()

                # advance epoch
                epoch_num += 1

            # Done with this configuration, so we run the test set
            run_test_epoch(worker_num, epoch_num, loaders["test_loader"], resources)
            epoch_num += 1


def spawn_process_factory(training_configs: List[TrainingConfig],
                          core: CausalLMCore,
                          trainer: CausalLMTrainer,
                          optim_factory: Callable[[nn.Module], torch.optim.Optimizer],
                          schedule_factory: Callable[[torch.optim.Optimizer], torch.optim.lr_scheduler.LRScheduler],
                          ) -> Callable[[int], None]:

    """
    Provided with everything except the process number, and this
    will give you back something that can accept that process number
    in order to spawn a config bound to the device.

    :param training_configs:
    :param core:
    :param trainer:
    :param optim_factory:
    :param schedule_factory:
    :return:
    """
    template_config = training_configs[0]
    num_workers = template_config.num_workers
    def spawn_model(worker_num: int, core=core, trainer=trainer)->Tuple[CausalLMCore, CausalLMTrainer]:
        device = template_config.devices[worker_num]
        core = core.to(device)
        trainer = trainer.to(device)
        return core, trainer

    loader_factory = create_dataloader_factory(num_workers,
                                               core.vocabulary.tokenizer,
                                               template_config.loader_config)

    run_spawned_process = functools.partial(run_training_process,
                                            training_configs=training_configs,
                                            loader_factory=loader_factory,
                                            model_factory=spawn_model,
                                            optim_factory=optim_factory,
                                            schedule_factory=schedule_factory
                                            )
    return run_spawned_process






"""
Defines some default loss mechanisms
"""
from typing import Optional
from abc import ABC, abstractmethod

import torch
from torch import nn




class MainLossInterface(nn.Module):
    """
    Any main loss mechanism, which
    provides loss based on the model,
    must implement this
    """
    @abstractmethod
    def forward(self, logits: torch.Tensor, targets: torch.Tensor,  schedule: Optional[torch.Tensor])->torch.Tensor:
        """
        Definition for the main loss input and output
        :param logits: The logits feature to take loss with. Shape (..., d_logits)
        :param target: The target. Integers. Shape (...).
        :param schedule: An optional scheduling control feature. We can adjust loss strength here
        :return: The loss. A scalar.
        """

class MemAccessLossInterface(nn.Module):
    """
    The memory access loss mechanism
    must implement this interface.
    """
    @abstractmethod
    def forward(self, write_probability_mass: torch.Tensor, schedule: Optional[float])->torch.Tensor:
        """
        Definition for the memory access loss input and output
        :param write_probability_mass:
        - A feature indicating, per memory slot, how much we wrote to that slot
        - Shape (..., num_memories)
        :param schedule: An optional scheduling float
        :return: A scalar loss.
        """
class CrossEntropyLoss(MainLossInterface):
    """
    A fairly normal cross entropy loss
    function. If the schedule parameter
    is defined, we multiply by it before
    returning.
    """
    def __init__(self,
                 padding_token_id: int,
                 weight: float = 1.0,
                 label_smoothing_rate: float = 0.1
                 ):
        """

        :param padding_token_id: Padding token to ignore
        :param weight: A static weight to scale by
        :param label_smoothing_rate: The label smoothing rate
        """
        super().__init__()
        self.weight = weight
        self.cross_entropy = nn.CrossEntropyLoss(ignore_index=padding_token_id,
                                                 label_smoothing=label_smoothing_rate,
                                                 reduction="sum")
    def forward(self, logits: torch.Tensor, targets: torch.Tensor, schedule: Optional[float]):
        loss = self.weight*self.cross_entropy(logits, targets)
        if schedule is not None:
            loss = loss*schedule
        return loss


def kl_divergence(p: torch.Tensor, q: torch.Tensor, epsilon=1e-10) -> torch.Tensor:
    """
    Computes the Kullback-Leibler divergence between two probability distributions P and Q.

    :param p: Tensor of the true distribution (should sum to 1 along the last dimension).
    :param q: Tensor of the approximate distribution (should sum to 1 along the last dimension).
    :param epsilon: Small constant to prevent division by zero and log of zero.
    :return: Tensor representing the KL divergence between P and Q.
    """
    # Ensure neither P nor Q has zero entries (prevents NaN in log and division)
    p = p.clamp(min=epsilon)
    q = q.clamp(min=epsilon)

    # Compute KL divergence
    kl_div = p * torch.log(p / q)

    # Sum over the last dimension
    return kl_div.sum(dim=-1)


class UniformMemLoss(MemAccessLossInterface):
    """
    Encourages the memory access pattern to occur uniformly,
    with memory write rates being even across all slots.
    """

    def __init__(self, weight: float = 1.0):
        super().__init__()
        self.weight = weight

    def forward(self, write_probability_mass: torch.Tensor, schedule: Optional[float] = None) -> torch.Tensor:
        """
        Defines the memory access loss to promote uniform memory access.

        :param write_probability_mass: A tensor indicating how much was written to each memory slot.
                                       Shape (..., num_memories)
        :param schedule: An optional scheduling scalar.
        :return: A scalar loss.
        """
        # Normalize the write probability mass into a probability distribution
        write_probability_mass = write_probability_mass / (write_probability_mass.sum(dim=-1, keepdim=True) + 1e-4)

        # Compute KL divergence against a uniform target distribution
        loss = self.weight * self.kl_divergence(write_probability_mass)

        # Apply scheduling if provided
        if schedule is not None:
            loss = loss * schedule
        return loss

    def kl_divergence(self, write_prob: torch.Tensor) -> torch.Tensor:
        """
        Helper method to compute KL divergence between the write distribution
        and a uniform distribution.

        :param write_prob: The write probability distribution tensor.
        :return: KL divergence loss.
        """
        # Calculate log of the write probability distribution
        write_log_predictions = torch.log(write_prob + 1e-4)

        # Create the uniform target distribution
        num_memories = write_prob.shape[-1]
        target_distribution = torch.full_like(write_prob, 1 / num_memories)

        # Compute KL divergence
        kl_divergence = write_prob * (write_log_predictions - torch.log(target_distribution))
        return kl_divergence.sum(dim=-1).mean()  # Sum over the last dimension and average



import os
import textwrap
from typing import Dict, Union, Tuple

import torch
from torch import nn
from transformers import (PreTrainedModel, AutoTokenizer, PreTrainedTokenizer,
                          AutoModelForCausalLM, PreTrainedTokenizerFast)
from tokenizers import processors
from enum import Enum

class SpecialTokens(Enum):
    bos_token = "<BOS>"
    eos_token = "<EOS>"
    pad_token = "<PAD>"

class AdditionalSpecialTokens(Enum):
    prompt_token = "<PROMPT>" # Prompt tokens are tokens the model cannot generate itself
    beginning_of_response_token = "<RESPONSE>" # These the model are now responsible for generating
class Vocabulary(nn.Module):
    """
    A centralized place in which vocabulary,
    embeddings, and logits can be kept and
    otherwise managed.
    """
    def save_pretrained_vocabulary(self, directory: Union[str, os.PathLike]):
        """
        Saves the vocabulary in its current configuration
        :param directory: The directory to save in.
        """
        self.tokenizer.save_pretrained(directory)
        torch.save(self.embeddings, os.path.join(directory, "embeddings.pt"))
        torch.save(self.logit_projector, os.path.join(directory, "logit_projector.pt"))

    @classmethod
    def load_pretrained_vocabulary(cls, directory: Union[str, os.PathLike]) -> 'Vocabulary':
        """
        Loads a pretrained vocabulary from the indicated save directory
        :param directory: The directory to save in
        :return: The pretrained vocabulary struct
        """
        tokenizer = AutoTokenizer.from_pretrained(directory)
        embeddings = torch.load(os.path.join(directory, "embeddings.pt"))
        logit_projector = torch.load(os.path.join(directory, "logit_projector.pt"))
        return cls(embeddings, tokenizer, logit_projector)

    def __reduce__(self):
        msg = """
        Attempt was made to invoke torch.save or pickle directly. This 
        is not supported. Instead, use built in save methods on the classes.
        """
        msg = textwrap.dedent(msg)
        raise NotImplementedError(msg)

    def __init__(self,
                 embeddings: nn.Embedding,
                 tokenizer: PreTrainedTokenizer,
                 logit: nn.Linear,
                 ):
        super().__init__()

        assert embeddings.weight.shape[0] == logit.weight.shape[0], "Embeddings and logits did not have same model dim"

        self.d_model = embeddings.weight.shape[-1]
        self.embeddings = embeddings
        self.tokenizer = tokenizer
        self.logit_projector = logit

TokenizerAlias = Union[PreTrainedTokenizer, PreTrainedTokenizerFast]
def _customize_tokenizer(tokenizer: TokenizerAlias)->Tuple[TokenizerAlias, int]:
    """
    Customizes the provided pretrained tokenizer to include, possibly, extra special
    tokens for various situations. This will include injecting additional special
    tokens for the prompt-response logic, eos, bos, and pad conditions. It also will
    include inserting

    :param tokenizer: The tokenizer to customize
    :return:
    - The tokenizer
    - The number of new tokens
    """
    # Generate the update dictionary, and integrate it
    # into the tokenizer
    original_size = tokenizer.vocab_size

    # Setup special token update. Integrate them
    special_tokens = {case.name: case.value for case in SpecialTokens}
    special_tokens["additional_special_tokens"] = [case.value for case in AdditionalSpecialTokens]
    num_added_tokens = tokenizer.add_special_tokens(special_tokens, replace_additional_special_tokens=False)

    special_token_ids = special_tokens.copy()
    additional_special_tokens = special_token_ids.pop("additional_special_tokens")
    special_token_ids = [(value, tokenizer.convert_tokens_to_ids(value)) for value in special_token_ids.values()]
    special_token_ids += [(value, tokenizer.convert_tokens_to_ids(value)) for value in additional_special_tokens]

    # Define the postprocessor.
    #
    # The post processor is configured to either
    # inject SOS and EOS in single configuration,
    # or that PLUS prompt, response singles for
    # pair configuration.

    single_directive = [SpecialTokens.bos_token.value, "$A", SpecialTokens.eos_token.value]
    pair_directive = [SpecialTokens.bos_token.value,
                      AdditionalSpecialTokens.prompt_token.value,
                      "$A",
                      AdditionalSpecialTokens.beginning_of_response_token.value,
                      "$B",
                       SpecialTokens.eos_token.value
                       ]

    single_directive = " ".join(single_directive)
    pair_directive = " ".join(pair_directive)

    postprocessor = processors.TemplateProcessing(
        single =single_directive,
        pair=pair_directive,
        special_tokens=special_token_ids
    )
    if isinstance(tokenizer, PreTrainedTokenizerFast):
        tokenizer._tokenizer.post_processor = postprocessor
    elif isinstance(tokenizer, PreTrainedTokenizer):
        tokenizer.post_processor = postprocessor
    else:
        raise TypeError()

    tokenizer.true_vocab_size = original_size + num_added_tokens
    return tokenizer, num_added_tokens
def _load_tokenizer_from_huggingface(name: str, kwargs) -> Tuple[TokenizerAlias, int]:
    """
    Loads, and customizes as needed, the tokenizer vocabulary to be compatible
    with the model architecture. This consists of

    - Updating any missing special tokens to be present within the vocabulary
    - Updating the post processor to handle said tokens as needed


    :param name: The name of the model to load the head off of.
    :return:
    - The number of additional tokens that had to be injected
    - The tokenizer. Ready to tokenize. Note that some addition
    """
    # Ready the tokenizer, by inserting custom
    # vocabulary
    tokenizer = AutoTokenizer.from_pretrained(name, **kwargs)
    tokenizer, num_new_vocab_elements = _customize_tokenizer(tokenizer)
    return tokenizer, num_new_vocab_elements

def _load_embeddings_for_tokenizer(model: nn.Module, num_added_tokens: int)->nn.Embedding:
    """
    Loads embeddings off of a module, and modifies them to be
    compatible with the tokenizer.
    :param model: Huggingface causal model to load off
    :param num_tokens_to_expand: The number of tokens to expand by
    :return: The setup embeddings.
    """
    with torch.no_grad():
        embeddings = model.get_input_embeddings()
        assert embeddings is not None

        old_num_embeddings, embedding_dim = embeddings.weight.size()
        new_embeddings = nn.Embedding(old_num_embeddings + num_added_tokens, embedding_dim)
        new_embeddings.weight[:old_num_embeddings] = embeddings.weight
    return new_embeddings

def _load_logits_for_tokenizer(model: nn.Module, num_added_tokens: int)->nn.Linear:
    """
    Loads the logits for the model, and expands them to be compatible
    based on the given number of additional tokens

    :param model: The huggingface model to load from
    :param num_added_tokens: The additional tokens
    :return: The logit layer
    """
    # Expand the logits to be able to predict the extra dimensions.
    with torch.no_grad():
        logits = model.get_output_embeddings()
        assert logits is not None

        num_logits, d_embedding = logits.weight.size()
        if logits.bias is None:
            new_logits = nn.Linear(d_embedding, num_logits + num_added_tokens, bias=False)
            new_logits.weight[:num_logits] = logits.weight
        else:
            new_logits = nn.Linear(d_embedding, num_logits)
            new_logits.weight[:num_logits] = logits.weight
            new_logits.bias[:num_logits] = logits.bias
    return new_logits
def load_vocabulary_off_huggingface_model(name: str, **kwargs) -> 'Vocabulary':
    """
    Loads, then customizes, a causal lm vocabulary for (hopefully)
    compatibility with the model.

    ---- requirements ----

    To construct a vocabulary struct, we need to get three things.
    These are.

    - A functional tokenizer with the right customizations.
    - An embedding that can accept the provided tokens
    - A logit projector that can predict the tokens

    --- General logic---

    Basically, we fetch the

    In particular, we fetch three
    things. These are

    - A tokenizer.

    :param name: The model to load from. We are going to
    :return: The vocabulary, configured for use.
    """

    tokenizer, num_new_vocab_elements = _load_tokenizer_from_huggingface(name, kwargs)
    donor_model = AutoModelForCausalLM.from_pretrained(name)
    embeddings = _load_embeddings_for_tokenizer(donor_model, num_new_vocab_elements)
    logits = _load_logits_for_tokenizer(donor_model, num_new_vocab_elements)
    return Vocabulary(embeddings, tokenizer, logits)


import functools
import torch
import datasets
import numpy as np
from dataclasses import dataclass, field
from datasets import DatasetDict, Dataset, load_dataset
from transformers import PreTrainedTokenizer, PreTrainedTokenizerFast
from torch.utils import data
from typing import Dict, List, Tuple, Callable, Union, Optional, Any
from .base import make_buffered_pipeline

@dataclass
class PretrainingLoaderConfig:
    """
    A configuration class for holding configuration
    details for a pretraining loader, sans external
    dependencies
    """
    huggingface_dataset_name: str
    huggingface_dataset_version: str
    batch_size: int
    truncate_length: int
    shuffle: bool = True
    num_batches_in_buffer: int = 20
    num_prefetch_threads: int = 4
    prefetch_factor = 4
    loader_kwargs: Dict[str, Any] = field(default_factory= lambda : {})\

def create_pretokenized_datasets(datasets: DatasetDict,
                                 tokenizer: PreTrainedTokenizer,
                                 use_cache: bool = True,
                                 ) -> DatasetDict:
    """
    Creates a pretokenized datasets collection
    :param datasets: A huggingface datasets collection of text.
    :param tokenizer: The tokenizer to use for this process
    :param use_cache: Whether to use the pretokenized cache if available, or force a rebuild
    :return: The pretokenized dataset.
    """

    def batch_tokenize(examples: Dataset):
        texts = examples["text"]
        encodings = tokenizer(
            text=texts,
            truncation=False,
            padding=False,
            return_attention_mask=False,
        )
        return encodings

    datasets = datasets.map(batch_tokenize, batched=True, remove_columns="text", load_from_cache_file=use_cache)
    return datasets


def data_collator(batch: List[Dict[str, List[int]]],
                  pad_id: int,
                  truncate_length: int,
                  ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    A function designed to be bound to using functools.

    It will finish padding and tokenization.

    :param batch: The batch under consideration
    :param pad_id: The integer to use when padding.
    :param truncate_length: What to truncate if we go beyond
    :return: The tokens, targets, and attn mask.
    """
    batch_ids = [item["input_ids"] for item in batch]

    # Figure out the padding/truncate lengt
    target_length = max(len(item) for item in batch_ids)
    target_length = min(truncate_length, target_length)

    # Perform the padding/truncate process
    final_ids = []
    final_masks = []
    for case in batch_ids:
        # Encode the mask and inputs array. Make sure
        # to truncate if needed.
        ids_array = torch.tensor(case[:target_length], dtype=torch.long)
        mask_array = torch.ones_like(ids_array, dtype=torch.bool)

        # Integrate the padding to whatever degree is needed
        ids_padding = torch.full([target_length - len(ids_array)], pad_id, dtype=torch.long)
        mask_padding = torch.zeros_like(ids_padding, dtype=torch.bool)

        # Combine and store
        ids_output = torch.concat([ids_array, ids_padding], dim=-1)
        attn_mask = torch.concat([mask_array, mask_padding], dim=-1)

        final_ids.append(ids_output)
        final_masks.append(attn_mask)

    input_ids = torch.stack(final_ids, dim=0)
    attn_mask = torch.stack(final_masks, dim=0)

    # Produce the targets and inputs
    inputs = input_ids[..., :-1]
    targets = input_ids[..., 1:]
    batch_mask = attn_mask[..., :-1].to(dtype=torch.bool)

    # Pin

    inputs.pin_memory()
    targets.pin_memory()
    batch_mask.pin_memory()

    # Return
    return inputs, targets, batch_mask


def make_dataloaders(worker_rank: int,
                     total_workers: int,
                     config: PretrainingLoaderConfig,
                     padding_id: int,
                     datasets: DatasetDict
                     ) -> Dict[str, data.DataLoader]:
    """
    Creates the dataloaders for the given dataset dictionaries.

    :param worker_rank: The rank of the worker under consideration
    :param total_workers: The total number of workers
    :param config: The config we are working
    :param padding_id: The integer to use when padding.
    :param datasets: The datasets to convert
    :return: The dataloaders
    """

    bound_collate_fn = functools.partial(data_collator, pad_id=padding_id, truncate_length=config.truncate_length)
    loaders: Dict[str, data.DataLoader] = {}
    for name, dataset in datasets.items():
        loader = make_buffered_pipeline(config.batch_size,
                                        total_workers,
                                        worker_rank,
                                        bound_collate_fn,
                                        dataset,
                                        num_batches_in_buffer=config.num_batches_in_buffer,
                                        shuffle=config.shuffle,
                                        num_prefetch_threads=config.num_prefetch_threads,
                                        prefetch_factor=config.prefetch_factor
                                        )
        loaders[name] = loader
    return loaders


def create_dataloader_factory(total_workers: int,
                              tokenizer: PreTrainedTokenizer,
                              config: PretrainingLoaderConfig
                              ) -> Callable[[int], Dict[str, data.DataLoader]]:
    """
    Creates a multiprocessing factory that is designed to produce
    data loader factories compatible with a particular device
    :param total_workers: The total number of workers in the env
    :param tokenizer: The tokenizer to use
    :param config: The loader config to use
    :return: The loaders. 'train_loader', 'test_loader', 'validation_loader'
    """

    # Setup the tokenized datasets
    raw_dataset = load_dataset(config.huggingface_dataset_name,
                               config.huggingface_dataset_version,
                               **config.loader_kwargs)
    pretokenized_dataset = create_pretokenized_datasets(raw_dataset, tokenizer)

    # Bind the factory. It now is looking only for the
    # worker number in order to finish initializing.

    return functools.partial(make_dataloaders,
                             total_workers=total_workers,
                             config=config,
                             padding_id=tokenizer.pad_token_id,
                             datasets=pretokenized_dataset
                             )



from typing import Dict, Any

import csv
from typing import List, Dict, Any


class LogMetrics:
    def __init__(self, file: str, metric_slots: List[str]):
        """
        A logging interface for saving metrics to a CSV file.

        :param file: The file path to save to.
        :param metric_slots: List of metric names to be tracked.
        """
        self.file = file
        self.metric_slots = ["epoch", "batch", "worker"] + metric_slots
        self._initialized = False  # To check if headers are written

    def __call__(self, epoch: int, batch: int, worker: int, metrics: Dict[str, Any]):
        """
        Logs metrics for a given epoch, batch, and worker.

        :param epoch: Current epoch number.
        :param batch: Current batch number.
        :param worker: Worker ID or number.
        :param metrics: Dictionary of metric names and their values.
        """
        # Ensure all metric slots are present in the data
        row = {
            "epoch": epoch,
            "batch": batch,
            "worker": worker,
            **metrics  # Merges the metrics dictionary into the row
        }

        # Write to file
        with open(self.file, 'a', newline='') as f:
            writer = csv.DictWriter(f, fieldnames=self.metric_slots)

            # Write headers if this is the first write
            if not self._initialized:
                writer.writeheader()
                self._initialized = True

            # Write the row data
            writer.writerow(row)



import numpy as np
import torch
from torch.nn import functional as F
from torch.utils import data
from torch.utils.data import Dataset, IterableDataset
from typing import Iterable, List, Callable, Any
from collections.abc import Sized


class NumpyBufferedStream:
    """
    A specialized buffer that can perform
    a vectorized pop, push, and length process.

    It will express the length of the items in
    its buffer, automatically refill itself when
    possible if items are popped, and allows a vectorized
    pop process.
    """

    def pop(self, indices: np.ndarray) -> np.ndarray:
        """
        Pops the given indices out of the array,
        returning them in the specified order. Additional
        elements are then loaded into the buffer if
        possible to bring it up to capacity.

        :param indices: The indices to pop.
        :return: The popped array.
        """
        chosen_mask = np.zeros(len(self.buffer), dtype=bool)
        chosen_mask[indices] = True
        output = self.buffer[chosen_mask]
        self.buffer = self.buffer[~chosen_mask]
        self.refill_buffer()
        return output

    def get_lengths(self) -> np.ndarray:
        """
        Gets the length of each item in the buffer,
        and returns it as an array.
        """
        return np.array([len(item) for item in self.buffer])

    def refill_buffer(self):
        """
        Refill the buffer by loading batches from the stream.
        """
        # Load a batch from the stream, appending it to the buffer
        try:
            while len(self.buffer) < self.buffer_size:
                batch = next(self.stream)
                # Flatten batch to 1D array if needed and append to the buffer
                batch = np.array(batch, dtype=object)
                self.buffer = np.concatenate([self.buffer, batch])
        except StopIteration:
            # Stop if the stream is exhausted
            pass

    def is_buffer_empty(self) -> bool:
        """Returns True if the buffer is empty."""
        return len(self.buffer) == 0

    def __init__(self,
                 stream: Iterable[Sized],
                 buffer_size: int):
        """
        Initialize the NumpyBufferedStream.

        :param stream: A batched iterable where each item is a batch of data.
        :param buffer_size: The maximum size of the buffer.
        """
        self.stream = iter(stream)
        self.buffer = np.array([], dtype=object)
        self.buffer_size = buffer_size
        self.refill_buffer()


class BufferedBatchSampler(data.Sampler):
    """
    A Buffered batch sampler, responsible
    for sampling from an entire batch with
    a buffer for like sized collating
    """

    @staticmethod
    def select_batch_indices_using_clustoid(batch_size: int, lengths: np.ndarray) -> np.ndarray:
        """
        Selects the most effective batch. This will be the
        one where the average distance between elements is lowest
        over batch_size.

        :param batch_size: The size of the batch to form.
        :param lengths: The lengths associated with each potential batch in the cache.
        :return: The indices associated with the batch.
        """
        differences = lengths[:, None] - lengths[None, :]
        distances = np.abs(differences)

        indexes = np.argsort(distances, axis=-1)[:, :batch_size]
        distances = np.take_along_axis(distances, indexes, axis=-1)
        batch_distances = np.sum(distances, axis=-1)
        best_centroid_index = np.argmin(batch_distances)

        return indexes[best_centroid_index]

    def __init__(self,
                 batch_size: int,
                 stream_head: data.DataLoader,
                 num_batches_in_buffer: int = 20,
                 ):
        super().__init__(stream_head)
        self.buffer_stream_head = stream_head
        self.buffer_size = batch_size * num_batches_in_buffer
        self.batch_size = batch_size

    def __len__(self) -> int:
        return len(self.buffer_stream_head)

    def __iter__(self):
        buffer = NumpyBufferedStream(self.buffer_stream_head, self.buffer_size)
        while not buffer.is_buffer_empty():
            # Get lengths for clustering
            lengths = buffer.get_lengths()
            batch_indices = self.select_batch_indices_using_clustoid(self.batch_size, lengths)

            # Create the batch by popping selected items from the buffer
            batch = buffer.pop(batch_indices)
            yield batch


def make_buffered_pipeline(batch_size: int,
                           num_workers: int,
                           worker_rank: int,
                           collate_fn: Callable[[List[int]], torch.Tensor],
                           pretokenized_dataset: Dataset,
                           num_batches_in_buffer: int = 20,
                           shuffle: bool = True,
                           num_prefetch_threads: int = 4,
                           prefetch_factor: int = 2
                           ) -> data.DataLoader:
    """
    Creates a buffered pipeline out of the
    given parameters

    :param batch_size: Size of batch
    :param num_workers: Total number of workers
    :param worker_rank: The rank of this worker
    :param collate_fn: The batch collator. Will accept pretokenized inputs
    :param pretokenized_dataset: The pretokenized dataset
    :param num_batches_in_buffer: The number of batches in the buffer
    :param shuffle: Whether to shuffle each iteration
    :param num_prefetch_threads: The number of prefetch threads
    :param prefetch_factor: The number of batches to prefetch
    :return: The dataloader
    """

    # Create the primary loader. This will be responsible
    # for shuffling and accommodation of distribute processes,
    # and getting data off the hard drive.
    distributed_sampler = data.DistributedSampler(pretokenized_dataset,
                                                  num_replicas=num_workers,
                                                  rank=worker_rank,
                                                  shuffle=shuffle)

    stream_loader = data.DataLoader(pretokenized_dataset,
                                    batch_size=batch_size,
                                    shuffle=False,
                                    collate_fn=lambda x : x,
                                    sampler=distributed_sampler)

    # Create the buffered loader. This will accept the primary
    # loader, and loads a certain number of the cases that it
    # keeps in a buffer. It chooses to batch cases that
    # have similar lengths

    batch_sampler = BufferedBatchSampler(batch_size, stream_loader, num_batches_in_buffer)

    return data.DataLoader(pretokenized_dataset,
                           batch_sampler=batch_sampler,
                           num_workers=num_prefetch_threads,
                           prefetch_factor=prefetch_factor,
                           collate_fn=collate_fn,
                           pin_memory=True)


